from enum import Enum
from typing import Protocol

import httpx
import typer
from rich import print
from rich.console import Console
from rich.table import Column
from rich.table import Table
from rich.text import Text

from nstimes.departure import Departure
from nstimes.styles import cancelled
from nstimes.styles import cyan
from nstimes.styles import green
from nstimes.styles import red


class Printer(Protocol):
    title: str = ""

    def generate_output(self) -> None:
        """generates output in the console"""

    def add_departure(self, departure: Departure) -> None:
        """adds a row to the departures"""


class ConsolePrinter:
    def __init__(self) -> None:
        self.buf = ""
        self.title = ""
        self.lines: list[str] = []

    def generate_output(self) -> None:
        print(self.title)
        print("\n")
        for line in self.lines:
            print(line)

    def add_departure(self, departure: Departure) -> None:
        line = f"{departure.train_type:<3s} p.{departure.platform:>3s} in {departure.calc_time_left_minutes():>2d} min {departure.departure_time} -> {departure.arrival_time}"
        if departure.cancelled:
            line = cancelled(line)
        self.lines.append(line)


class ConsoleTablePrinter:
    def __init__(self) -> None:
        self.table = Table(
            Column("Train", justify="left"),
            Column("Platform", justify="right"),
            Column("Leaves in", justify="right"),
            Column("Departure time", justify="right"),
            Column("Arrival time", justify="right"),
        )

    def generate_output(self) -> None:
        Console().print(self.table)

    @property
    def title(self) -> str:
        return str(self.table.title)

    @title.setter
    def title(self, value: str) -> None:
        self.table.title = value

    def add_departure(self, departure: Departure) -> None:
        if departure.cancelled:
            self.table.add_row(
                cancelled(departure.train_type),
                cancelled(departure.platform),
                cancelled(f"{departure.time_left_minutes} min"),
                cancelled(str(departure.departure_time)),
                cancelled(str(departure.arrival_time)),
            )
        else:
            self.table.add_row(
                departure.train_type,
                cyan(departure.platform),
                f"{cyan(departure.time_left_minutes)} min",
                f"{departure.departure_time}",
                f"{departure.arrival_time}",
            )


class PrinterChoice(str, Enum):
    table = "table"
    ascii = "ascii"


def get_printer(
    printer_choice: PrinterChoice,
) -> Printer:
    if printer_choice == PrinterChoice.ascii:
        return ConsolePrinter()
    elif printer_choice == PrinterChoice.table:
        return ConsoleTablePrinter()
    else:
        raise typer.Exit(1)



import json
import os
from datetime import datetime

import httpx
import typer


SCRIPT_DIR = os.path.dirname(__file__)
STATIONS_FILE = os.path.join(SCRIPT_DIR, "stations.json")
DATE_FORMAT = "%d-%m-%Y"
TIME_FORMAT = "%H:%M"


def get_uic_mapping() -> dict[str, str]:
    with open(STATIONS_FILE, "r", encoding="utf-8") as file:
        uic_mapping: dict[str, str] = json.load(file)
    return uic_mapping


def get_headers(token: str) -> dict[str, str]:
    return {"Cache-Control": "no-cache", "Ocp-Apim-Subscription-Key": token}


def httpx_get(token: str, query_params: dict[str, str], api: str) -> httpx.Response:
    API_URL = "https://gateway.apiportal.ns.nl/reisinformatie-api/api"
    try:
        with httpx.Client() as client:
            response = client.get(
                url=f"{API_URL}/{api}", headers=get_headers(token), params=query_params
            )
            response.raise_for_status()
            return response
    except httpx.ReadTimeout:
        raise typer.Exit(2)
    except httpx.HTTPStatusError:
        raise typer.Exit(1)


def convert_to_rfc3339(time: str, date: str) -> str:
    datetime_obj = datetime.strptime(f"{date} {time}", f"{DATE_FORMAT} {TIME_FORMAT}")
    rfc3339_str = datetime_obj.isoformat()
    return rfc3339_str



from dataclasses import dataclass  # pragma: no cover
from datetime import datetime
from typing import Optional

from pydantic import computed_field

from nstimes.styles import green
from nstimes.styles import red
from nstimes.utils import get_uic_mapping
from nstimes.utils import httpx_get

DATETIME_FORMAT_STRING = "%Y-%m-%dT%H:%M:%S%z"


class Time:
    def __init__(self, planned: datetime, actual: Optional[datetime] = None) -> None:
        self.actual = actual or planned
        self.planned = planned

    @property
    def delay_minutes(self) -> int:
        delay = self.actual - self.planned
        return int(delay.total_seconds() / 60)

    def __str__(self) -> str:
        time_str = self.planned.strftime("%H:%M")
        delay_str = "" if self.delay_minutes == 0 else red(f"+{self.delay_minutes}")
        return f"({time_str}{delay_str})"


@dataclass
class Departure:
    train_type: str
    platform: str
    departure_time: Time
    arrival_time: Time
    cancelled: bool = False

    @computed_field  # type: ignore
    @property
    def time_left_minutes(self) -> int:
        return self.calc_time_left_minutes()

    def calc_time_left_minutes(self, reference_time: datetime = datetime.now()) -> int:
        reference_time = reference_time.replace(
            tzinfo=self.departure_time.actual.tzinfo
        )
        time_left = self.departure_time.actual - reference_time
        return int(time_left.total_seconds() / 60)


def parse_time(data: dict) -> Time:  # type: ignore
    planned_time = datetime.strptime(data["plannedDateTime"], DATETIME_FORMAT_STRING)
    actual_time = data.get("actualDateTime")
    if actual_time is not None:
        actual_time = datetime.strptime(actual_time, DATETIME_FORMAT_STRING)
    return Time(planned=planned_time, actual=actual_time)


def get_departures(
    start: str,
    end: str,
    token: str,
    rdc3339_datetime: str,
    max_len: Optional[int] = None,
) -> list[Departure]:
    uic_mapping = get_uic_mapping()

    query_params = {
        "originUicCode": uic_mapping[start],
        "destinationUicCode": uic_mapping[end],
        "dateTime": rdc3339_datetime,
    }
    response = httpx_get(token=token, query_params=query_params, api="v3/trips")
    departures = []
    trips = response.json()["trips"]
    for trip in trips:
        trip = trip["legs"][0]
        origin = trip["origin"]
        destination = trip["destination"]
        track = origin.get("actualTrack", origin.get("plannedTrack", "?"))

        cancelled = trip.get("cancelled", False)
        departure_time = parse_time(origin)
        arrival_time = parse_time(destination)

        train_type = trip["product"]["categoryCode"]

        departure = Departure(
            train_type=train_type,
            platform=track,
            departure_time=departure_time,
            arrival_time=arrival_time,
            cancelled=cancelled,
        )
        if departure.time_left_minutes < 0:
            continue
        departures.append(departure)
    return departures



def red(text: str | int) -> str:
    return f"[bold red]{text}[/bold red]"


def cyan(text: str | int) -> str:
    return f"[bold cyan]{text}[/bold cyan]"


def green(text: str | int) -> str:
    return f"[bold green]{text}[/bold green]"


def strike(text: str | int) -> str:
    return f"[strike]{text}[/strike]"


def cancelled(text: str | int) -> str:
    return red(strike(text))



import json
import os
from datetime import datetime
from importlib.metadata import version
from typing import Generator

import httpx
import typer
from dotenv import load_dotenv
from typing_extensions import Annotated

from nstimes.departure import get_departures
from nstimes.printers import get_printer
from nstimes.printers import PrinterChoice
from nstimes.utils import convert_to_rfc3339
from nstimes.utils import DATE_FORMAT
from nstimes.utils import get_uic_mapping
from nstimes.utils import httpx_get
from nstimes.utils import STATIONS_FILE
from nstimes.utils import TIME_FORMAT

# Load environment variables from the .env file
load_dotenv()


app = typer.Typer(
    help="Find your next train home while you are in CLI. I used the Dutch Railway Services (Nederlandse Spoorwegen) API to make myself this tool.",
    pretty_exceptions_show_locals=os.getenv("SHOW_LOCALS"),
)


@app.command(help="Generate stations lookup, should not be neccesary", hidden=True)
def update_stations_json(
    token: Annotated[
        str, typer.Option(help="Token to talk with the NS API", envvar="NS_API_TOKEN")
    ],
    path: Annotated[str, typer.Option(help="Path for stations.json")] = STATIONS_FILE,
) -> None:
    query_params = {"countryCodes": "nl"}
    response = httpx_get(token=token, query_params=query_params, api="v2/stations")
    # get list of stations to uic code
    data = response.json()["payload"]
    uic_mapping = {d["namen"]["lang"]: d["UICCode"] for d in data}
    with open(path, "w", encoding="utf-8") as file:
        json.dump(uic_mapping, file)
    typer.Exit(0)


def complete_name(lut: dict[str, str], incomplete: str) -> Generator[str, None, None]:
    for name in lut:
        if name.startswith(incomplete):
            yield name


complete_station_name = lambda incomplete: complete_name(
    lut=get_uic_mapping(), incomplete=incomplete
)


@app.command(
    help="Provide train type, platform and departure times of an A -> B journey"
)
def journey(
    start: Annotated[
        str, typer.Option(help="Start station", autocompletion=complete_station_name)
    ],
    end: Annotated[
        str, typer.Option(help="Stop station", autocompletion=complete_station_name)
    ],
    token: Annotated[
        str, typer.Option(help="Token to talk with the NS API", envvar="NS_API_TOKEN")
    ],
    time: Annotated[
        str, typer.Option(help=f"Time to departure ({TIME_FORMAT})")
    ] = datetime.now().strftime("%H:%M"),
    date: Annotated[
        str, typer.Option(help=f"Date to departure ({DATE_FORMAT})")
    ] = datetime.now().strftime("%d-%m-%Y"),
    printer_choice: Annotated[
        PrinterChoice, typer.Option(help="The type of printer")
    ] = PrinterChoice.ascii,
) -> None:
    printer = get_printer(printer_choice)

    printer.title = f"Journeys from {start} -> {end} at {date} {time}"
    rdc3339_datetime = convert_to_rfc3339(time, date)
    try:
        departures = get_departures(
            start=start, end=end, token=token, rdc3339_datetime=rdc3339_datetime
        )
    except KeyError:
        print("One or more of the stations does not exist")
        raise typer.Exit(1)
    for departure in departures:
        if departure.time_left_minutes >= 0:
            printer.add_departure(departure)
    printer.generate_output()


def version_callback(value: bool) -> None:
    if value:
        print(f"nstimes version: {version(__package__)}")
        raise typer.Exit(0)


@app.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print version info",
    ),
) -> None:
    return


if __name__ == "__main__":
    app()  # pragma: no cover
